from os.path import join as join_path
from sqlite3 import Connection
from sqlite3 import Cursor
from sqlite3 import connect
from typing import List
from typing import Union


def tiered_path(id_: Union[int, str], depth: int = 5, width: int = 2) -> str:
    assert isinstance(id_, int) or (isinstance(id_, str) and id_.isdigit())
    assert isinstance(depth, int) and depth > 0
    assert isinstance(width, int) and width > 0

    id_str: str = str(int(id_)).zfill(depth * width)
    return join_path(*[id_str[n:n + width] for n in range(0, depth * width, width)])


def connect_database(path: str) -> Connection:
    return connect(path)


def insert(db: Connection, table: str, keys: List[str], values: List[Union[int, str]], replace: bool = True):
    db.execute(
        f"""INSERT OR {"REPLACE" if replace else "IGNORE"} INTO {table}
        ({",".join(keys)})
        VALUES ({",".join(["?"] * len(values))})""",
        values
    )


def select(db: Connection, table: str, fields: List[str], keys: List[str], key_values: List[Union[int, str]],
           like: bool = False, and_: bool = True, order: List[str] = None
           ) -> Cursor:
    op: str = "like" if like else "="
    logic: str = "AND" if and_ else "OR"
    where: List[str] = [f"{key} {op} ?" for key in keys]
    order = [] if order is None else order
    return db.execute(
        f'''SELECT {",".join(fields)} FROM {table}
        WHERE {f" {logic} ".join(where)}
        {"ORDER BY " + ','.join(order) if order else ""}''',
        key_values
    )


def select_all(db: Connection, table: str, fields: List[str], order: List[str] = None) -> Cursor:
    order = [] if order is None else order
    return db.execute(f'''SELECT {",".join(fields)} FROM {table} {"ORDER BY " + ','.join(order) if order else ""}''')


def update(db: Connection, table: str, fields: List[str], values: List[Union[int, str]], key: str, key_value: str):
    assert len(fields) == len(values) and len(fields) > 0

    update_values: List[str] = [f"{u} = ?" for u in fields]

    db.execute(
        f"""UPDATE {table}
        SET {",".join(update_values)}
        WHERE {key} = ?""",
        (*values, key_value,)
    )


def delete(db: Connection, table: str, key: str, key_value: Union[int, str]):
    db.execute(f"""DELETE FROM {table} where {key} = ?""", (key_value,))


def count(db: Connection, table: str) -> int:
    return db.execute(f"SELECT COUNT(*) FROM {table}").fetchall()[0][0]


def vacuum(db: Connection):
    db.execute("VACUUM")


def check_errors(db: Connection, table: str) -> List[tuple]:
    if (table := table.upper()) in ("SUBMISSIONS", "JOURNALS"):
        errors: List[tuple] = []
        errors.extend(select(db, table, ["*"], "ID", 0).fetchall())
        errors.extend(select(db, table, ["*"], "AUTHOR", "").fetchall())
        errors.extend(select(db, table, ["*"], "TITLE", "").fetchall())
        errors.extend(select(db, table, ["*"], "UDATE", "").fetchall())

        if table == "SUBMISSIONS":
            errors.extend(select(db, table, ["*"], "FILELINK", "").fetchall())
            errors.extend(select(db, table, ["*"], "FILESAVED", "").fetchall())

        return sorted(set(errors), key=lambda s: s[0])
    else:
        return []


def make_tables(db: Connection):
    from .journals import make_journals_table
    from .settings import make_settings_table
    from .submissions import make_submissions_table
    from .users import make_users_table

    make_journals_table(db)
    make_settings_table(db)
    make_submissions_table(db)
    make_users_table(db)

    db.commit()



from os import makedirs
from os.path import join as join_path
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from filetype import guess_extension

from .database import Connection
from .database import delete
from .database import insert
from .database import select
from .database import tiered_path
from .settings import read_setting

"""
Entries guide - SUBMISSIONS
v3.2
0  ID
1  AUTHOR
2  TITLE
3  UDATE
4  DESCRIPTION
5  TAGS
6  CATEGORY
7  SPECIES
8  GENDER
9  RATING
10 FILELINK
11 FILEEXT
12 FILESAVED
13 LOCATION
14 SERVER
"""

submissions_table: str = "SUBMISSIONS"
submissions_fields: List[str] = [
    "ID", "AUTHOR", "TITLE",
    "UDATE", "DESCRIPTION", "TAGS",
    "CATEGORY", "SPECIES", "GENDER",
    "RATING", "FILELINK", "FILEEXT",
    "FILESAVED"
]
submissions_indexes: Dict[str, int] = {f: i for i, f in enumerate(submissions_fields)}


def make_submissions_table(db: Connection):
    db.execute(
        f"""CREATE TABLE IF NOT EXISTS {submissions_table}
        (ID INT UNIQUE NOT NULL,
        AUTHOR TEXT NOT NULL,
        TITLE TEXT,
        UDATE DATE NOT NULL,
        DESCRIPTION TEXT,
        TAGS TEXT,
        CATEGORY TEXT,
        SPECIES TEXT,
        GENDER TEXT,
        RATING TEXT,
        FILELINK TEXT,
        FILEEXT TEXT,
        FILESAVED INT,
        PRIMARY KEY (ID ASC));"""
    )


def exist_submission(db: Connection, submission_id: int) -> bool:
    return bool(select(db, submissions_table, ["ID"], ["ID"], [submission_id]).fetchone())


def remove_submission(db: Connection, submission_id: int):
    delete(db, submissions_table, "ID", submission_id)
    db.commit()


def save_submission(db: Connection, submission: Dict[str, Union[str, int]], file: Optional[bytes]):
    file_ext: str = ""
    file_url_name: str = submission["file_url"].split("/")[-1]
    file_url_ext: str = file_url_name.split(".")[-1] if "." in file_url_name else ""

    if file:
        if (file_ext_tmp := guess_extension(file)) is None:
            file_ext = file_url_ext
        elif (file_ext := str(file_ext_tmp)) == "zip" and file_url_ext:
            file_ext = file_url_ext

        sub_folder: str = join_path(read_setting(db, "FILESFOLDER"), tiered_path(submission["id"]))
        sub_file: str = "submission" + (f".{file_ext}" if file_ext else "")

        makedirs(sub_folder, exist_ok=True)

        with open(join_path(sub_folder, sub_file), "wb") as f:
            f.write(file)

    insert(db, submissions_table,
           list(submissions_indexes.keys()),
           [submission["id"], submission["author"], submission["title"],
            submission["date"], submission["description"], ",".join(sorted(submission["tags"], key=str.lower)),
            submission["category"], submission["species"], submission["gender"],
            submission["rating"], submission["file_url"], file_ext,
            file is not None],
           replace=True)

    db.commit()


def search_submissions(db: Connection,
                       author: List[str] = None, title: List[str] = None,
                       date: List[str] = None, description: List[str] = None, tags: List[str] = None,
                       category: List[str] = None, species: List[str] = None, gender: List[str] = None,
                       rating: List[str] = None, limit: List[Union[str, int]] = None,
                       offset: List[Union[str, int]] = None, order: List[str] = None
                       ) -> List[tuple]:
    author = [] if author is None else list(map(str.lower, author))
    title = [] if title is None else list(map(str.lower, title))
    date = [] if date is None else list(map(str.lower, date))
    description = [] if description is None else list(map(str.lower, description))
    tags = [] if tags is None else list(map(str.lower, tags))
    category = [] if category is None else list(map(str.lower, category))
    species = [] if species is None else list(map(str.lower, species))
    gender = [] if gender is None else list(map(str.lower, gender))
    rating = [] if rating is None else list(map(str.lower, rating))

    assert any((author, title, date, description, tags, category, species, gender, rating))

    wheres: List[str] = [
        " OR ".join(["UDATE like ?"] * len(date)),
        " OR ".join(["lower(RATING) like ?"] * len(rating)),
        " OR ".join(["lower(GENDER) like ?"] * len(gender)),
        " OR ".join(["lower(SPECIES) like ?"] * len(species)),
        " OR ".join(["lower(CATEGORY) like ?"] * len(category)),
        " OR ".join(['replace(lower(AUTHOR), "_", "") like ?'] * len(author)),
        " OR ".join(["lower(TITLE) like ?"] * len(title)),
        " OR ".join(["lower(TAGS) like ?"] * len(tags)),
        " OR ".join(["lower(DESCRIPTION) like ?"] * len(description))
    ]

    wheres_str: str = " AND ".join(map(lambda p: "(" + p + ")", filter(len, wheres)))
    order_str: str = f"ORDER BY {','.join(order)}" if order else ""
    limit_str: str = f"LIMIT {int(limit[0])}" if limit is not None else ""
    offset_str: str = f"OFFSET {int(offset[0])}" if offset is not None else ""

    return db.execute(
        f"""SELECT * FROM {submissions_table} WHERE {wheres_str} {order_str} {limit_str} {offset_str}""",
        date + rating + gender + species + category + author + title + tags + description
    ).fetchall()



from .__version__ import __version__
from .database import check_errors
from .database import connect_database
from .database import connect_database
from .database import count
from .database import delete
from .database import insert
from .database import make_tables
from .database import select
from .database import select_all
from .database import tiered_path
from .database import update
from .database import vacuum
from .journals import exist_journal
from .journals import journals_fields
from .journals import journals_indexes
from .journals import journals_table
from .journals import remove_journal
from .journals import save_journal
from .journals import search_journals
from .settings import add_history
from .settings import read_history
from .settings import read_setting
from .settings import settings_table
from .settings import write_setting
from .submissions import exist_submission
from .submissions import remove_submission
from .submissions import save_submission
from .submissions import search_submissions
from .submissions import submissions_fields
from .submissions import submissions_indexes
from .submissions import submissions_table
from .update import update_database
from .users import edit_user_field_add
from .users import edit_user_field_remove
from .users import edit_user_field_replace
from .users import edit_user_remove_journal
from .users import edit_user_remove_submission
from .users import exist_user
from .users import exist_user_field_value
from .users import find_user_from_fields
from .users import find_user_from_galleries
from .users import find_user_from_journal
from .users import find_user_from_submission
from .users import new_user
from .users import remove_user
from .users import search_users
from .users import users_fields
from .users import users_indexes
from .users import users_table
